from datetime import datetime

from pytz import timezone


def tznow(tzname):
    """
    Get the current ``datetime`` in the specified timezone.

    Args:
        tzname: the ISO standard timezone name,
        such as ``Africa/Cairo``, ``America/Los_Angeles``.
    """
    return datetime.now(timezone(tzname))


def pacificnow():
    return tznow("America/Los_Angeles")


def make_timestamp() -> str:
    """
    This function creates a timestamp string with fixed format like

        '2020-08-22T08:09:13.401346'

    Strings created by this function can be compared to
    determine time order. There is no need to parse the string
    into `datetime` objects.

    The returned string is often written as a timestamp file, like

        open(file_name, 'w').write(make_timestamp())
    """
    return datetime.utcnow().isoformat(timespec="microseconds")



import asyncio
import concurrent.futures
import functools
import logging

logger = logging.getLogger(__name__)


async def async_call(
    func, *args, executor: concurrent.futures.Executor = None, **kwargs
):
    loop = asyncio.get_running_loop()
    if kwargs:
        func = functools.partial(func, **kwargs)

    return await loop.run_in_executor(executor, func, *args)



import sys
import time
from contextlib import contextmanager
from functools import wraps
from typing import Callable, List


def humanize(seconds: float) -> List[str]:
    msg = []
    if seconds >= 3600:
        hours, seconds = divmod(seconds, 3600)
        hours = int(hours)
        if hours < 2:
            msg.append(f"{hours} hour")
        else:
            msg.append(f"{hours} hours")
    if seconds >= 60:
        minutes, seconds = divmod(seconds, 60)
        minutes = int(minutes)
        if minutes < 2:
            msg.append(f"{minutes} minute")
        else:
            msg.append(f"{minutes} minutes")
    msg.append(f"{round(seconds, 4)} seconds")
    return msg


def timed(print_func: Callable = None) -> Callable:
    """
    Usage 1:

        @timed()
        def myfunc(...):
            ...

    Usage 2:

        import logging
        logger = logging.getLogger(__name__)

        @timed(logger.info)
        def myfunc(...):
            ...
    """
    if print_func is None:

        def print_func(text):
            print(text, file=sys.stderr)

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def profiled_func(*args, **kwargs):
            print_func(f"Starting function `{func.__name__}`")
            t0 = time.monotonic()
            result = func(*args, **kwargs)
            duration = ", ".join(humanize(time.monotonic() - t0))
            print_func(f"Finishing function `{func.__name__}`")
            print_func(f"Function `{func.__name__}` took {duration} to finish")
            return result

        return profiled_func

    return decorator


def timed_call(func, *args, **kwargs):
    t0 = time.monotonic()
    z = func(*args, **kwargs)
    seconds = time.monotonic() - t0
    return z, seconds


@contextmanager
def timer(msg: str):
    """
    Time a code block:

        with timer('my block'):
            x = 3
            y = 4
            ...
    """
    t0 = time.monotonic()
    yield
    t1 = time.monotonic()
    print(f"{msg}: {humanize(t1 - t0)}")



import subprocess


def get_host_ip():
    """
    Get the IP address of the host machine from within a Docker container.
    """
    # INTERNAL_HOST_IP=$(ip route show default | awk '/default/ {print $3})')
    # another idea:
    # ip -4 route list match 0/0 | cut -d' ' -f3
    #
    # Usually the result is '172.17.0.1'
    # This is the "loopback" address used in code to connect to the host machine.
    # In most documentation this is said to be '127.0.0.1'.

    z = subprocess.check_output(["ip", "-4", "route", "list", "match", "0/0"])
    z = z.decode()[len("default via ") :]
    return z[: z.find(" ")]



"""
This is a small Python language exercise.
Recommendation for practical use of this decorator is not implied.
"""

import functools
from typing import Any, Callable


class Infix:
    """
    Suppose we have

        @Infix
        def func(x, y):
            return ...

    Then,

        x |func| y

    is equivalent to

        func(x, y)
    """

    def __init__(self, function: Callable[[Any, Any], Any]):
        self.function = function

    def __ror__(self, other):
        return Infix(functools.partial(self.function, other))

    def __or__(self, other):
        return self.function(other)

    def __call__(self, x, y):
        return self.function(x, y)


infix = Infix



import cProfile
import pstats
import warnings
from functools import wraps
from io import StringIO
from typing import Callable


def profiled(
    top: int = 32, sort_by: str = None, prof_file: str = None
) -> Callable[[Callable], Callable]:
    """
    Print out profiling info for a test function,
    and optionally dump the profile result in specified file
    for later inspection.

    Args:
        top: number of top items to print.
        sort_by: list of sort keys for print-out, if not `None`.
            If `None` (the default), a default list of sort keys are used.
            If `[]`, no print-out.
        prof_file: file name for a dump of the profile.
            Use `snakeviz` to view the content of this file later.
            If `None`, a default file name is used.
            If `''`, profile is not saved.

    Usage::

        @profiled()
        def func():
            ...

        @profiled(top=20)
        def func():
            ...

    In a test function, ::

        def test_abc():
            @profiled()
            def do_it():
                ...

            do_it()

    """

    if sort_by is None:
        sort_by = ["cumtime", "tottime"]

    if isinstance(sort_by, str):
        sort_by = [sort_by]

    if prof_file is None:
        prof_file = "cprofile.out"

    def mydecorator(func):
        @wraps(func)
        def profiled_func(*args, **kwargs):
            profile = cProfile.Profile()
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()

            for sb in sort_by:
                s = StringIO()
                pstats.Stats(profile, stream=s).sort_stats(sb).print_stats(top)
                print("")
                print(s.getvalue())

            if prof_file:
                profile.dump_stats(prof_file)
                print("")
                print(
                    "profiling results are saved in",
                    prof_file,
                    "; view its content using `snakeviz`",
                )

            return result

        return profiled_func

    return mydecorator


try:
    import line_profiler
except ImportError:
    pass
else:

    def lineprofiled(
        *funcs,
    ) -> Callable[[Callable], Callable]:  # pylint: disable=unused-argument
        """
        A line-profiling decorator.

        Args:
            funcs: functions (function objects, not function names)
                to be line-profiled. If no function is specified,
                the function being decorated is profiled.

        Example:

                @lineprofiled()
                def myfunc(a, b):
                    g(a)
                    f(b)

                @lineprofiled(g, f)
                def yourfunc(a, b):
                    x = g(a)
                    y = f(b)
                    s(x, y)
        """

        def mydecorator(func):
            nonlocal funcs
            if not funcs:
                funcs = [func]

            @wraps(func)
            def profiled_func(*args, **kwargs):
                func_names = [f.__name__ for f in funcs]
                profile = line_profiler.LineProfiler(*funcs)
                z = profile.runcall(func, *args, **kwargs)

                profile.print_stats()
                stats = profile.get_stats()
                if not stats.timings:
                    warnings.warn("No profile stats.")
                    return z

                for key, timings in stats.timings.items():
                    if key[-1] in func_names:
                        if len(timings) > 0:
                            func_names.remove(key[-1])
                            if not func_names:
                                break
                if func_names:
                    # Force warnings.warn() to omit the source code line
                    # in the message
                    formatwarning_orig = warnings.formatwarning
                    warnings.formatwarning = lambda message, category, filename, lineno, line=None: formatwarning_orig(
                        message,
                        category,
                        filename,
                        lineno,
                        line="",
                    )
                    warnings.warn("No profile stats for %s." % str(func_names))
                    # Restore warning formatting.
                    warnings.formatwarning = formatwarning_orig

                return z

            return profiled_func

        return mydecorator



"""
Configure logging, mainly the format.

A call to the function ``config_logger`` in a launching script
is all that is needed to set up the logging format.
Usually the 'level' argument is the only argument one needs to customize::

  config_logger(level='info')

Do not call this in library modules. Library modules should have ::

   logger = logging.getLogger(__name__)

and then just use ``logger`` to write logs without concern about formatting,
destination of the log message, etc.
"""
import inspect
import logging
import logging.handlers
import os
import sys
import warnings
from datetime import datetime
from typing import Union

# When exceptions are raised during logging, then,
# the default implementation of handleError() in Handler
# checks to see if a module-level variable,
#   raiseExceptions,
# is set.
# If set, a traceback is printed to sys.stderr.
# If not set, the exception is swallowed.
#
# If no logging configuration is provided, then for Python 2.x,
#   If logging.raiseExceptions is False (production mode),
#       the event is silently dropped.
#   If logging.raiseExceptions is True (development mode),
#       a message 'No handlers could be found for logger X.Y.Z'
#       is printed once.


# Turn off annoyance in ptpython when setting DEBUG logging
logging.getLogger("parso").setLevel(logging.ERROR)

logging.captureWarnings(True)
warnings.filterwarnings("default", category=ResourceWarning)
warnings.filterwarnings("default", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning, module="ptpython")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="jedi")


rootlogger = logging.getLogger()
logger = logging.getLogger(__name__)


def get_calling_file():
    st = inspect.stack()
    caller = None
    for s in st:
        if os.path.basename(s.filename) == "runpy.py":
            break
        if "_pytest/python.py" in s.filename:
            # Don't follow into py.test; stop at the test file
            break
        if s.filename == "<stdin>":
            break
        caller = s
    return caller


def formatter(*, with_process_name: bool = False, with_thread_name: bool = False):
    tz = datetime.now().astimezone().tzname()
    msg = (
        "[%(asctime)s.%(msecs)03d "
        + tz
        + "; %(levelname)s; %(name)s, %(funcName)s, %(lineno)d]"
    )
    msg += "  "

    if with_process_name and with_thread_name:
        fmt = f"{msg}[%(processName)s (%(process)d) %(threadName)s]  %(message)s"
    elif with_process_name:
        fmt = f"{msg}[%(processName)s]  %(message)s"
    elif with_thread_name:
        fmt = f"{msg}[%(threadName)s]  %(message)s"
    else:
        fmt = f"{msg}%(message)s"

    return logging.Formatter(fmt=fmt, datefmt="%Y-%m-%d %H:%M:%S")


def set_level(level: Union[str, int] = logging.INFO):
    """
    In one application, call `set_level` on the top level once.
    Do not set level anywhere else, and do not set level on any handler.
    """
    if isinstance(level, str):
        level = getattr(logging, level.upper())
    level0 = rootlogger.level
    rootlogger.setLevel(level)
    return level0


def use_console_handler(**kwargs):
    if any(getattr(h, "_name_", None) == "console" for h in rootlogger.handlers):
        raise RuntimeError("the console handler is already in being used")
    h = logging.StreamHandler()
    h.setFormatter(formatter(**kwargs))
    h._name_ = "console"
    rootlogger.addHandler(h)


def _unuse_handler(name: str):
    for h in rootlogger.handlers:
        if getattr(h, "_name_", None) == name:
            rootlogger.removeHandler(h)
            return 1
    return 0


def unuse_console_handler():
    return _unuse_handler("console")


def use_disk_handler(
    *, foldername: str = None, maxBytes=1_000_000, backupCount=20, delay=True, **kwargs
):
    if any(getattr(h, "_name_", None) == "disk" for h in rootlogger.handlers):
        raise RuntimeError("the disk handler is already in being used")

    if foldername:
        foldername = foldername.rstrip("/")
    else:
        launcher = get_calling_file()
        foldername == f"{os.environ.get('LOGDIR', '/tmp/log')}/{launcher.lstrip('/').replace('/', '-')}"
    print(f"Log files are located in '{foldername}'")
    os.makedirs(foldername, exist_ok=True)
    h = logging.handlers.RotatingFileHandler(
        filename=foldername + "/current",
        maxBytes=maxBytes,
        backupCount=backupCount,
        delay=delay,
    )
    h.setFormatter(formatter(**kwargs))
    h._name_ = "disk"
    rootlogger.addHandler(h)
    return foldername


def unuse_disk_handler():
    return _unuse_handler("disk")


def log_uncaught_exception(handlers=None, logger=logger):
    # locally bind `handlers` and `logger` in case the global references are gone
    # when the exception handler is invoked.
    if handlers is None:
        handlers = rootlogger.handlers

    def handle_exception(exc_type, exc_val, exc_tb):
        if not issubclass(exc_tb, KeyboardInterrupt):
            if sys.version_info.minor < 11:
                logging.currentframe = lambda: sys._getframe(1)
            fn, lno, func, sinfo = logger.findCaller(stack_info=False, stacklevel=1)
            record = logger.makeRecord(
                logger.name,
                logging.CRITICAL,
                fn,
                lno,
                msg=exc_val,
                args=(),
                exc_info=(exc_type, exc_val, exc_tb),
                func=func,
                sinfo=sinfo,
            )

            for h in rootlogger.handlers:
                if getattr(h, "_name_", None) != "console":
                    h.handle(record)

        sys.__excepthook__(exc_type, exc_val, exc_tb)

    sys.excepthook = handle_exception


def config_logger(level=logging.INFO, **kwargs):
    # For use in one-off scripts.
    set_level(level)
    use_console_handler(**kwargs)
